# /* Copyright (C) 2001
#  * Housemarque Oy
#  * http://www.housemarque.com
#  *
#  * Distributed under the Boost Software License, Version 1.0. (See
#  * accompanying file LICENSE_1_0.txt or copy at
#  * http://www.boost.org/LICENSE_1_0.txt)
#  */
#
# /* Revised by Paul Mensonides (2002) */
# /* Revised by Edward Diener (2020) */
#
# /* See http://www.boost.org for most recent version. */
#
# ifndef BOOST_PREPROCESSOR_REPETITION_FOR_512_HPP
# define BOOST_PREPROCESSOR_REPETITION_FOR_512_HPP
#
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_257(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_258(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_259(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_260(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_261(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_262(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_263(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_264(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_265(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_266(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_267(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_268(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_269(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_270(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_271(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_272(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_273(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_274(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_275(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_276(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_277(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_278(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_279(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_280(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_281(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_282(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_283(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_284(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_285(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_286(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_287(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_288(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_289(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_290(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_291(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_292(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_293(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_294(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_295(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_296(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_297(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_298(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_299(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_300(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_301(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_302(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_303(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_304(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_305(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_306(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_307(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_308(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_309(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_310(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_311(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_312(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_313(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_314(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_315(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_316(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_317(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_318(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_319(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_320(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_321(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_322(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_323(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_324(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_325(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_326(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_327(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_328(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_329(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_330(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_331(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_332(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_333(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_334(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_335(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_336(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_337(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_338(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_339(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_340(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_341(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_342(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_343(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_344(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_345(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_346(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_347(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_348(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_349(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_350(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_351(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_352(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_353(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_354(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_355(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_356(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_357(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_358(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_359(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_360(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_361(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_362(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_363(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_364(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_365(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_366(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_367(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_368(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_369(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_370(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_371(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_372(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_373(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_374(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_375(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_376(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_377(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_378(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_379(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_380(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_381(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_382(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_383(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_384(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_385(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_386(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_387(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_388(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_389(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_390(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_391(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_392(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_393(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_394(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_395(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_396(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_397(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_398(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_399(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_400(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_401(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_402(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_403(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_404(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_405(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_406(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_407(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_408(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_409(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_410(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_411(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_412(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_413(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_414(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_415(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_416(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_417(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_418(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_419(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_420(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_421(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_422(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_423(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_424(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_425(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_426(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_427(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_428(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_429(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_430(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_431(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_432(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_433(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_434(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_435(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_436(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_437(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_438(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_439(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_440(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_441(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_442(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_443(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_444(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_445(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_446(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_447(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_448(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_449(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_450(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_451(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_452(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_453(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_454(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_455(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_456(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_457(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_458(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_459(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_460(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_461(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_462(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_463(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_464(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_465(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_466(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_467(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_468(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_469(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_470(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_471(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_472(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_473(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_474(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_475(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_476(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_477(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_478(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_479(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_480(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_481(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_482(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_483(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_484(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_485(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_486(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_487(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_488(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_489(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_490(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_491(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_492(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_493(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_494(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_495(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_496(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_497(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_498(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_499(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_500(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_501(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_502(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_503(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_504(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_505(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_506(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_507(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_508(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_509(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_510(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_511(s, p, o, m) 0
# define BOOST_PP_FOR_CHECK_BOOST_PP_FOR_512(s, p, o, m) 0
#
# endif
