//  Copyright (c) 2001-2011 Hartmut Kaiser
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#if !defined(BOOST_SPIRIT_KARMA_NONTERMINAL_MAR_05_2007_0539PM)
#define BOOST_SPIRIT_KARMA_NONTERMINAL_MAR_05_2007_0539PM

#if defined(_MSC_VER)
#pragma once
#endif

#include <boost/spirit/home/karma/nonterminal/rule.hpp>
#include <boost/spirit/home/karma/nonterminal/grammar.hpp>
#include <boost/spirit/home/karma/nonterminal/debug_handler.hpp>
#include <boost/spirit/home/karma/nonterminal/simple_trace.hpp>

#endif
