// size_t.h
// Copyright (c) 2007-2009 Ben Hanson (http://www.benhanson.net/)
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file licence_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
#ifndef BOOST_SPIRIT_SUPPORT_DETAIL_LEXER_SIZE_T_HPP
#define BOOST_SPIRIT_SUPPORT_DETAIL_LEXER_SIZE_T_HPP

#include <stddef.h> // ptrdiff_t

#include <string>

#endif
