//
// Copyright (c) 2022 Klemens Morgenstern (klemens.morgenstern@gmx.net)
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//

#ifndef BOOST_COBALT_HPP
#define BOOST_COBALT_HPP

#include <boost/cobalt/async_for.hpp>
#include <boost/cobalt/channel.hpp>
#include <boost/cobalt/concepts.hpp>
#include <boost/cobalt/config.hpp>
#include <boost/cobalt/detached.hpp>
#include <boost/cobalt/error.hpp>
#include <boost/cobalt/gather.hpp>
#include <boost/cobalt/generator.hpp>
#include <boost/cobalt/join.hpp>
#include <boost/cobalt/leaf.hpp>
#include <boost/cobalt/main.hpp>
#include <boost/cobalt/op.hpp>
#include <boost/cobalt/promise.hpp>
#include <boost/cobalt/run.hpp>
#include <boost/cobalt/race.hpp>
#include <boost/cobalt/spawn.hpp>
#include <boost/cobalt/task.hpp>
#include <boost/cobalt/this_coro.hpp>
#include <boost/cobalt/this_thread.hpp>
#include <boost/cobalt/thread.hpp>
#include <boost/cobalt/wait_group.hpp>
#include <boost/cobalt/with.hpp>

#endif //BOOST_COBALT_HPP

