//
// Copyright 2012 Christian Henning
//
// Distributed under the Boost Software License, Version 1.0
// See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt
//
#ifndef BOOST_GIL_EXTENSION_TOOLBOX_COLOR_SPACES_HPP
#define BOOST_GIL_EXTENSION_TOOLBOX_COLOR_SPACES_HPP

#include <boost/gil/extension/toolbox/color_spaces/cmyka.hpp>
#include <boost/gil/extension/toolbox/color_spaces/gray_alpha.hpp>
#include <boost/gil/extension/toolbox/color_spaces/hsl.hpp>
#include <boost/gil/extension/toolbox/color_spaces/hsv.hpp>
#include <boost/gil/extension/toolbox/color_spaces/lab.hpp>
#include <boost/gil/extension/toolbox/color_spaces/xyz.hpp>

#endif
