//
// Copyright (c) 2020 Alexander Grund
//
// Distributed under the Boost Software License, Version 1.0.
// https://www.boost.org/LICENSE_1_0.txt

#ifndef BOOST_NOWIDE_DETAIL_UTF_HPP_INCLUDED
#define BOOST_NOWIDE_DETAIL_UTF_HPP_INCLUDED

#include <boost/nowide/utf/utf.hpp>

// Legacy compatibility header only. Include <boost/nowide/utf/utf.hpp> instead

namespace boost {
namespace nowide {
    namespace detail {
        namespace utf = boost::nowide::utf;
    } // namespace detail
} // namespace nowide
} // namespace boost

#endif
